import six
from six.moves import range
import logging, time, json, socket
from collections import MutableMapping
import requests.exceptions as rexc
from .utils import urlpathjoin, xauth
from .serialization import jlencode, jldecode, mpdecode
from .serialization import MSGPACK_AVAILABLE

logger = logging.getLogger('hubstorage.resourcetype')
CHUNK_SIZE = 512


class ResourceType(object):

    resource_type = None
    key_suffix = None

    def __init__(self, client, key, auth=None):
        self.client = client
        self.key = urlpathjoin(self.resource_type, key, self.key_suffix)
        self.auth = xauth(auth) or client.auth
        self.url = urlpathjoin(client.endpoint, self.key)

    def _allows_mpack(self, path=None):
        """ Check if request can be served with msgpack data.

        Currently, items, logs, collections and samples endpoints are able to
        return msgpack data. However, /stats calls can only return JSON data
        for now.
        """
        if not MSGPACK_AVAILABLE or path == 'stats':
            return False
        return self.resource_type in ('items', 'logs',
                                      'collections', 'samples')

    @staticmethod
    def _enforce_msgpack(**kwargs):
        kwargs.setdefault('headers', {})
        kwargs['headers']['Accept'] = 'application/x-msgpack'
        return kwargs

    def _iter_content(self, _path, **kwargs):
        kwargs['url'] = urlpathjoin(self.url, _path)
        kwargs.setdefault('auth', self.auth)
        return self.client.request(**kwargs).iter_content(CHUNK_SIZE)

    def _iter_lines(self, _path, **kwargs):
        kwargs['url'] = urlpathjoin(self.url, _path)
        kwargs.setdefault('auth', self.auth)
        if 'jl' in kwargs:
            kwargs['data'] = jlencode(kwargs.pop('jl'))

        r = self.client.request(**kwargs)

        lines = r.iter_lines()
        if six.PY3:
            return (l.decode(r.encoding or 'utf8') for l in lines)
        return lines

    def apirequest(self, _path=None, **kwargs):
        if self._allows_mpack(_path) and kwargs.get('method').upper() == 'GET':
            kwargs = self._enforce_msgpack(**kwargs)
            return mpdecode(self._iter_content(_path=_path, **kwargs))
        return jldecode(self._iter_lines(_path, **kwargs))

    def apipost(self, _path=None, **kwargs):
        return self.apirequest(_path, method='POST', **kwargs)

    def apiget(self, _path=None, **kwargs):
        kwargs.setdefault('is_idempotent', True)
        return self.apirequest(_path, method='GET', **kwargs)

    def apidelete(self, _path=None, **kwargs):
        kwargs.setdefault('is_idempotent', True)
        return self.apirequest(_path, method='DELETE', **kwargs)


class DownloadableResource(ResourceType):
    MAX_RETRIES = 180
    RETRY_INTERVAL = 60

    @staticmethod
    def _add_key_meta(params):
        """Adds meta=_key to ensure a key is returned"""
        meta = params.get('meta', [])
        if '_key' not in meta:
            meta = list(meta)
            meta.append('_key')
            params['meta'] = meta
        return params

    def _add_resume_param(self, lastline, offset, params):
        """Adds a startafter=LASTKEY parameter if there was a lastvalue"""
        if lastline is not None:
            lastvalue = json.loads(lastline)
            params['startafter'] = lastvalue['_key']
            if 'start' in params:
                del params['start']

    def iter_values(self, *args, **kwargs):
        """Reliably iterate through all data as python objects

        calls either iter_json or iter_msgpack, decoding the results
        """
        if self._allows_mpack():
            return mpdecode(self.iter_msgpack(*args, **kwargs))
        return jldecode(self.iter_json(*args, **kwargs))

    def _retry(self, iter_callback, resume=False, _path=None, requests_params=None, **apiparams):
        """Reliable iterate through all data calling iter_callback"""
        self._add_key_meta(apiparams)
        lastexc = None
        chunk = None
        offset = 0
        for attempt in range(self.MAX_RETRIES):
            if resume:
                self._add_resume_param(chunk, offset, apiparams)
            try:
                for chunk in iter_callback(_path=_path, params=apiparams,
                                           **requests_params):
                    yield chunk
                    offset += 1
                break
            except (ValueError, socket.error, rexc.RequestException) as exc:
                # catch requests exceptions other than HTTPError
                if isinstance(exc, rexc.HTTPError):
                    raise
                lastexc = exc
                url = urlpathjoin(self.url, _path)
                msg = "Retrying read of %s in %ds: attempt=%d/%d error=%s"
                args = url, self.RETRY_INTERVAL, attempt, self.MAX_RETRIES, exc
                logger.debug(msg, *args)
                time.sleep(self.RETRY_INTERVAL)
        else:
            url = urlpathjoin(self.url, _path)
            logger.error("Failed %d times reading items from %s, params %s, "
                         "last error was: %s", self.MAX_RETRIES, url,
                         apiparams, lastexc)

    def iter_msgpack(self, _path=None, requests_params=None, **apiparams):
        """Reliably iterate through all data as msgpack"""
        requests_params = dict(requests_params or {})
        requests_params.setdefault('method', 'GET')
        requests_params.setdefault('stream', True)
        requests_params = self._enforce_msgpack(**requests_params)
        for chunk in self._retry(self._iter_content, False, _path,
                                 requests_params, **apiparams):
            yield chunk

    def iter_json(self, _path=None, requests_params=None, **apiparams):
        """Reliably iterate through all data as json strings"""
        requests_params = dict(requests_params or {})
        requests_params.setdefault('method', 'GET')
        requests_params.setdefault('stream', True)
        for line in self._retry(self._iter_lines, True, _path, requests_params,
                                **apiparams):
            yield line


class ItemsResourceType(ResourceType):

    batch_size = 1000
    batch_qsize = None  # defaults to twice batch_size if None
    batch_start = 0
    batch_interval = 15.0
    batch_content_encoding = 'identity'

    # batch writer reference in case of used
    _writer = None

    # TODO override _add_resume_param - can avoid requestomg _key by
    # deriving from project, spider, job and offset

    def batch_write_start(self):
        """Override to set a start parameter when commencing writing"""
        return 0

    @property
    def writer(self):
        if self._writer is None:
            self._writer = self.client.batchuploader.create_writer(
                url=self.url,
                auth=self.auth,
                size=self.batch_size,
                start=self.batch_write_start(),
                interval=self.batch_interval,
                qsize=self.batch_qsize,
                content_encoding=self.batch_content_encoding
            )
        return self._writer

    def flush(self):
        if self._writer is not None:
            self._writer.flush()

    def close(self, block=True):
        if self._writer is not None:
            self._writer.close(block=block)

    def write(self, item):
        return self.writer.write(item)

    def list(self, _key=None, **params):
        return self.apiget(_key, params=params)

    def get(self, _key, **params):
        """Return first matching result"""
        for o in self.list(_key, params=params):
            return o

    def stats(self):
        return next(self.apiget('stats'))


class MappingResourceType(ResourceType, MutableMapping):

    _cached = None
    ignore_fields = ()

    def __init__(self, *a, **kw):
        self._cached = kw.pop('cached', None)
        self._deleted = set()
        super(MappingResourceType, self).__init__(*a, **kw)

    def __str__(self):
        return str(self._data)

    def __repr__(self):
        return '{}({})'.format(self.__class__.__name__, repr(self._data))

    @property
    def _data(self):
        if self._cached is None:
            r = self.apiget()
            try:
                self._cached = next(r)
            except StopIteration:
                self._cached = {}

        return self._cached

    def expire(self):
        self._cached = None

    def save(self):
        for key in self._deleted:
            self.apidelete(key)
        self._deleted.clear()
        if self._cached:
            if not self.ignore_fields:
                self.apipost(jl=self._data, is_idempotent=True)
            else:
                self.apipost(jl={k: v for k, v in six.iteritems(self._data)
                                 if k not in self.ignore_fields},
                             is_idempotent=True)

    def __getitem__(self, key):
        return self._data[key]

    def __setitem__(self, key, value):
        self._data[key] = value
        self._deleted.discard(key)

    def __delitem__(self, key):
        del self._data[key]
        self._deleted.add(key)

    def __iter__(self):
        return iter(self._data)

    def __len__(self):
        return len(self._data)

    def liveget(self, key):
        for o in self.apiget(key):
            return o



"""
Test Client
"""
from .hstestcase import HSTestCase
from hubstorage import HubstorageClient
from hubstorage.utils import apipoll

class ClientTest(HSTestCase):

    def test_default_ua(self):
        self.assertEqual(self.hsclient.user_agent,
                         HubstorageClient.DEFAULT_USER_AGENT)

    def test_custom_ua(self):
        client = HubstorageClient(auth=HSTestCase.auth,
                                  endpoint=HSTestCase.endpoint,
                                  user_agent='testUA')
        self.assertEqual(client.user_agent, 'testUA')

    def test_push_job(self):
        c = self.hsclient
        c.push_job(self.projectid, self.spidername,
                   priority=self.project.jobq.PRIO_LOW,
                   foo='baz')
        job = self.start_job()
        m = job.metadata
        self.assertEqual(m.get('state'), u'running', c.auth)
        self.assertEqual(m.get('foo'), u'baz')
        self.project.jobq.finish(job)
        self.project.jobq.delete(job)

        # job auth token is valid only while job is running
        m = c.get_job(job.key).metadata
        self.assertEqual(m.get('state'), u'deleted')
        self.assertEqual(m.get('foo'), u'baz')

    def test_jobsummaries(self):
        hsc = self.hsclient
        # add at least one running or pending job to ensure summary is returned
        hsc.push_job(self.projectid, self.spidername, state='running')

        def _get_summary():
            jss = hsc.projects.jobsummaries()
            mjss = dict((str(js['project']), js) for js in jss)
            return mjss.get(self.projectid)
        summary = apipoll(_get_summary)
        self.assertIsNotNone(summary)

    def test_timestamp(self):
        ts1 = self.hsclient.server_timestamp()
        ts2 = self.hsclient.server_timestamp()
        self.assertGreater(ts1, 0)
        self.assertLessEqual(ts1, ts2)



"""
Test Project
"""
import json
import six
from six.moves import range
from random import randint, random
from requests.exceptions import HTTPError
from hubstorage import HubstorageClient
from hubstorage.utils import millitime
from .hstestcase import HSTestCase
from .testutil import failing_downloader


class ProjectTest(HSTestCase):

    def test_projectid(self):
        p1 = self.hsclient.get_project(int(self.projectid))
        p2 = self.hsclient.get_project(str(self.projectid))
        self.assertEqual(p1.projectid, p2.projectid)
        self.assertEqual(type(p1.projectid), six.text_type)
        self.assertEqual(type(p2.projectid), six.text_type)
        self.assertRaises(AssertionError, self.hsclient.get_project, '111/3')

    def test_get_job_from_key(self):
        job = self.project.push_job(self.spidername)
        parts = tuple(job.key.split('/'))
        self.assertEqual(len(parts), 3)
        self.assertEqual(parts[:2], (self.projectid, self.spiderid))
        samejob1 = self.hsclient.get_job(job.key)
        samejob2 = self.project.get_job(job.key)
        samejob3 = self.project.get_job(parts[1:])
        self.assertEqual(samejob1.key, job.key)
        self.assertEqual(samejob2.key, job.key)
        self.assertEqual(samejob3.key, job.key)

    def test_get_jobs(self):
        p = self.project
        j1 = p.push_job(self.spidername, testid=0)
        j2 = p.push_job(self.spidername, testid=1)
        j3 = p.push_job(self.spidername, testid=2)
        # global list must list at least one job
        self.assertTrue(list(p.get_jobs(count=1, state='pending')))
        # List all jobs for test spider
        r = list(p.get_jobs(spider=self.spidername, state='pending'))
        self.assertEqual([j.key for j in r], [j3.key, j2.key, j1.key])

    def test_get_jobs_with_legacy_filter(self):
        p = self.project
        j1 = p.push_job(self.spidername, state='finished',
                        close_reason='finished', tags=['t2'])
        j2 = p.push_job(self.spidername, state='finished',
                        close_reason='finished', tags=['t1'])
        j3 = p.push_job(self.spidername, state='pending')
        j4 = p.push_job(self.spidername, state='finished',
                        close_reason='failed', tags=['t1'])
        j5 = p.push_job(self.spidername + 'skip', state='finished',
                        close_reason='failed', tags=['t1'])

        filters = [['spider', '=', [self.spidername]],
                   ['state', '=', ['finished']],
                   ['close_reason', '=', ['finished']],
                   ['tags', 'haselement', ['t1']],
                   ['tags', 'hasnotelement', ['t2']]]
        jobs = p.get_jobs(filter=[json.dumps(x) for x in filters])
        assert [j.key for j in jobs] == [j2.key], jobs

    def test_push_job(self):
        job = self.project.push_job(self.spidername, state='running',
                                   priority=self.project.jobq.PRIO_HIGH,
                                   foo=u'bar')
        self.assertEqual(job.metadata.get('state'), u'running')
        self.assertEqual(job.metadata.get('foo'), u'bar')
        self.project.jobq.finish(job)
        self.project.jobq.delete(job)
        job.metadata.expire()
        self.assertEqual(job.metadata.get('state'), u'deleted')
        self.assertEqual(job.metadata.get('foo'), u'bar')

    def test_auth(self):
        # client without global auth set
        hsc = HubstorageClient(endpoint=self.hsclient.endpoint)
        self.assertEqual(hsc.auth, None)

        # check no-auth access
        try:
            hsc.push_job(self.projectid, self.spidername)
        except HTTPError as exc:
            self.assertTrue(exc.response.status_code, 401)
        else:
            self.assertTrue(False, '401 not raised')

        try:
            hsc.get_project(self.projectid).push_job(self.spidername)
        except HTTPError as exc:
            self.assertTrue(exc.response.status_code, 401)
        else:
            self.assertTrue(False, '401 not raised')

        try:
            hsc.get_job((self.projectid, 1, 1)).items.list()
        except HTTPError as exc:
            self.assertTrue(exc.response.status_code, 401)
        else:
            self.assertTrue(False, '401 not raised')

        try:
            hsc.get_project(self.projectid).get_job((self.projectid, 1, 1)).items.list()
        except HTTPError as exc:
            self.assertTrue(exc.response.status_code, 401)
        else:
            self.assertTrue(False, '401 not raised')

        # create project with auth
        auth = self.hsclient.auth
        project = hsc.get_project(self.projectid, auth)
        self.assertEqual(project.auth, auth)
        job = project.push_job(self.spidername)
        samejob = project.get_job(job.key)
        self.assertEqual(samejob.key, job.key)

    def test_broad(self):
        project = self.hsclient.get_project(self.projectid)
        # populate project with at least one job
        job = project.push_job(self.spidername)
        self.assertEqual(job.metadata.get('state'), 'pending')
        job = self.start_job()
        self.assertEqual(job.metadata.get('state'), 'running')
        job.items.write({'title': 'bar'})
        job.logs.info('nice to meet you')
        job.samples.write([1, 2, 3])
        job.close_writers()
        job.jobq.finish(job)

        # keep a jobid for get_job and unreference job
        jobid = job.key
        jobauth = job.auth
        del job

        self.assertTrue(list(project.items.list(self.spiderid, count=1)))
        self.assertTrue(list(project.logs.list(self.spiderid, count=1)))
        self.assertTrue(list(project.samples.list(self.spiderid, count=1)))

        job = project.client.get_job(jobid, jobauth=jobauth)
        job.purged()

    def test_settings(self):
        project = self.hsclient.get_project(self.projectid)
        settings = dict(project.settings)
        settings.pop('botgroups', None)  # ignore testsuite botgroups
        self.assertEqual(settings, {})
        project.settings['created'] = created = millitime()
        project.settings['botgroups'] = ['g1']
        project.settings.save()
        self.assertEqual(project.settings.liveget('created'), created)
        self.assertEqual(project.settings.liveget('botgroups'), ['g1'])
        project.settings.expire()
        self.assertEqual(dict(project.settings), {
            'created': created,
            'botgroups': ['g1'],
        })

    def test_requests(self):
        ts = millitime()
        job = self.project.push_job(self.spidername, state='running')
        # top parent
        r1 = job.requests.add(url='http://test.com/', status=200, method='GET',
                              rs=1337, duration=5, parent=None, ts=ts)
        # first child
        r2 = job.requests.add(url='http://test.com/2', status=400, method='POST',
                              rs=0, duration=1, parent=r1, ts=ts + 1)
        # another child with fingerprint set
        r3 = job.requests.add(url='http://test.com/3', status=400, method='PUT',
                              rs=0, duration=1, parent=r1, ts=ts + 2, fp='1234')

        job.requests.close()
        rr = job.requests.list()
        self.assertEqual(next(rr),
                         {u'status': 200, u'rs': 1337,
                          u'url': u'http://test.com/', u'time': ts,
                          u'duration': 5, u'method': u'GET'})
        self.assertEqual(next(rr),
                         {u'status': 400, u'parent': 0, u'rs': 0,
                          u'url': u'http://test.com/2', u'time': ts + 1,
                          u'duration': 1, u'method': u'POST'})
        self.assertEqual(next(rr),
                         {u'status': 400, u'fp': u'1234', u'parent': 0,
                          u'rs': 0, u'url': u'http://test.com/3',
                          u'time': ts + 2, u'duration': 1,
                          u'method': u'PUT'})

        self.assertRaises(StopIteration, next, rr)

    def test_samples(self):
        # no samples stored
        j1 = self.project.push_job(self.spidername, state='running')
        self.assertEqual(list(j1.samples.list()), [])
        # simple fill
        ts = millitime()
        j1.samples.write([ts, 1, 2, 3])
        j1.samples.write([ts + 1, 5, 9, 4])
        j1.samples.flush()
        o = list(j1.samples.list())
        self.assertEqual(len(o), 2)
        self.assertEqual(o[0], [ts, 1, 2, 3])
        self.assertEqual(o[1], [ts + 1, 5, 9, 4])

        # random fill
        j2 = self.project.push_job(self.spidername, state='running')
        samples = []
        ts = millitime()
        count = int(j2.samples.batch_size * (random() + randint(1, 5)))
        for _ in range(count):
            ts += randint(1, 2**16)
            row = [ts] + list(randint(0, 2**16) for _ in range(randint(0, 100)))
            samples.append(row)
            j2.samples.write(row)
        j2.samples.flush()
        o = list(j2.samples.list())
        self.assertEqual(len(o), count)
        for r1, r2 in zip(samples, o):
            self.assertEqual(r1, r2)

    def test_jobsummary(self):
        js = self.project.jobsummary()
        self.assertEqual(js.get('project'), int(self.project.projectid), js)
        self.assertEqual(js.get('has_capacity'), True, js)
        self.assertTrue('pending' in js, js)
        self.assertTrue('running' in js, js)

    def test_bulkdata(self):
        j = self.project.push_job(self.spidername, state='running')
        for i in range(20):
            j.logs.info("log line %d" % i)
            j.items.write(dict(field1="item%d" % i))
            j.requests.add("http://test.com/%d" % i,
                200, 'GET', 10, None, 10, 120)
        for resourcename in ('logs', 'items', 'requests'):
            resource = getattr(j, resourcename)
            resource.flush()

            # downloading resource, with simulated failures
            with failing_downloader(resource):
                downloaded = list(resource.iter_values())
                self.assertEqual(len(downloaded), 20)

    def test_output_string(self):
        project = self.hsclient.get_project(self.projectid)
        project.push_job(self.spidername)
        job = self.start_job()
        job.items.write({'foo': 'bar'})
        job.close_writers()
        items = self.hsclient.get_job(job.key).items.iter_json()
        self.assertEqual(type(next(items)), str)
